/* $Id: control.c,v 1.6 2000/11/27 20:09:44 malekith Exp $ */

#include "compiler.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* expr.c */
void compile_exp(CTX *ctx, struct token *t, struct var_type *type);
void compute_logic(CTX *ctx, struct token *t);
void convert(CTX *ctx, struct token *t, struct var_type *from,
	     struct var_type *to);

/* control.c */
void compile_stmt(CTX *ctx, struct token *t);

extern int debug;

static void compile_do_while(CTX *ctx, struct token *t)
{
	int start, end, break_b;

	if (debug)
		printf("# do {\n");

	start = add_id(ctx);
	end = add_id(ctx);
	break_b = ctx->break_id;
	ctx->break_id = end;
	resolve(start);
	
	compile_stmt(ctx, t->right);

	if (debug)
		printf("# } while (expr)\n");
	
	compute_logic(ctx, t->left);
	
	if (debug)
		printf("# /* ->: do_while */\n");
		
	emit(ctx,
		op_btrue, start,
		0
	);

	if (debug)
		printf("# /* fin: do_while */\n");
		
	resolve(end);
	ctx->break_id = break_b;
}

static void compile_while(CTX *ctx, struct token *t)
{
	int start, end;
	int break_b;

	if (debug)
		printf("# while (expr)\n");

	break_b = ctx->break_id;
	start = add_id(ctx);
	end = add_id(ctx);
	ctx->break_id = end;
	resolve(start);
	
	compute_logic(ctx, t->left);
	emit(ctx,
		op_bfalse, end,
		0
	);
	
	if (debug)
		printf("# { /* while */\n");
		
	compile_stmt(ctx, t->right);
	emit(ctx, 
		op_bra, start,
		0
	);

	if (debug)
		printf("# } /* while */\n");

	resolve(end);
	ctx->break_id = break_b;
}

static void compile_if(CTX *ctx, struct token *t)
{
	int id1, id2 = 0;
	struct token *e;

	if (debug)
		printf("# if (expr)\n");

	e = t->right;
	SYN(e, alternative);
	compute_logic(ctx, t->left);
	id1 = add_id(ctx);
	if (e->right)
		id2 = add_id(ctx);
	emit(ctx,
		op_bfalse, id1,
		0
	);
	if (debug)
		printf("# { /* if */\n");
	compile_stmt(ctx, e->left);
	if (e->right) {
		emit(ctx,
			op_bra, id2,
			0
		);
		if (debug)
			printf("# } else {\n");
		resolve(id1);
		compile_stmt(ctx, e->right);
		resolve(id2);
	} else 
		resolve(id1);
		
	if (debug)
		printf("# } /* if */\n");
}

static void init_var(CTX *ctx, struct var *v, struct token *t)
{
	struct token *exp, *name;
	
	if (t->right == 0)
		switch (v->type.id) {
		case sel_ptr:
			t->right = new_token(0, 0, sel_ptr, 0);
			break;
		case sel_int:
			t->right = new_token(0, 0, sel_int, 0);
			break;
		case sel_double:
			t->right = new_token(0, 0, sel_double, 0.0);
			break;
		case sel_string:
			t->right = new_token(0, 0, sel_string, "\"\"");
			break;
		default:
			halt();
		}

	name = new_token(0, 0, sel_id, v->name);
	exp = new_token(name, t->right, sel_op, '=');
	t->right = 0;
	
	name->line = exp->line = t->line;
	compile_exp(ctx, exp, 0);
	destroy_token(exp);
}

static void compile_localvar(CTX *ctx, struct token *type, struct token *t)
{
	struct var *v;
	
	assert(t->sel == sel_syn);
	if (t->v_int == syn_lvar_names) {
		compile_localvar(ctx, type, t->left);
		compile_localvar(ctx, type, t->right);
		return;
	}

	assert(t->v_int == syn_one_lvar);

	v = NEW(struct var);
	v->name = take_string(t->left);
	v->offset = ctx->sp;
	v->ctx = ctx_local;
	if (token2type(&v->type, type))
		croak(ctx);
	ctx->sp += size_of(&v->type);
	
	if (debug)
		printf("@ %s %s @%d;\n", name_of(&v->type),
			v->name, v->offset);

	if (ph_find(ctx->vars, v->name)) {
		oops(t, "redefintion of variable %s in %s::%s()", 
		     v->name, ctx->f->parent->name, ctx->f->name);
		croak(ctx);
	}
	
	ph_add(ctx->vars, v);

	init_var(ctx, v, t);
}

static void compile_return(CTX *ctx, struct token *t)
{
	struct var_type x;
	if ((ctx->f->type.id == sel_void && t->left != 0) ||
	    (ctx->f->type.id != sel_void && t->left == 0)) { 
		oops(t, "return with %svalue in function returing %svoid",
			t->left ? "" : "no ", t->left ? "" : "non-");
		croak(ctx);
	}

	if (t->left) {
		compile_exp(ctx, t->left, &x);
		convert(ctx, t, &x, &ctx->f->type);
	}
	emit(ctx, op_rts, 0);
}

void compile_stmt(CTX * ctx, struct token *t)
{
	/* FIXME: check_stack(ctx); */
	static int lev = 0;
	
	lev++;
	if (debug)
		printf("#\tclv = %d\n", lev);	
restart:
	if (t == 0)
		goto out;

	if (t->sel == sel_op && t->v_int == ';') {
		compile_stmt(ctx, t->left);
		/* this is to cut off recurence, to compile a 
		 * function in roighly speaking constant stack
		 * space, it was:
		 *      compile_stmt(ctx, t->right);
		 *      return;
		 */
		t = t->right;
		goto restart;
	} else if (t->sel == sel_syn)
		switch (t->v_int) {
		case syn_alternative:
			halt();
		case syn_if:
			compile_if(ctx, t);
			goto out;
		case syn_while:
			compile_while(ctx, t);
			goto out;
		case syn_do_while:
			compile_do_while(ctx, t);
			goto out;
		case syn_lvars:
			compile_localvar(ctx, t->left, t->right);
			goto out;
		case syn_return:
			compile_return(ctx, t);
			goto out;
		}

	compile_exp(ctx, t, 0);
out:
	lev--;
}
