/* $Id: op.h,v 1.7 2000/11/15 01:42:18 malekith Exp $ */

/* ultra-RISC virtual processor definition
 *
 * iptr = signed STH
 * sizeof(ptr) == sizeof(void*)!
 *
 * REGISTERS:
 *  %d		- data [int]
 *  %f		- data [float]
 *  %a		- address 
 */

enum {
	OOPS_ILL = 1,
	OOPS_HLT,
	OOPS_STACK,
	OOPS_DIV0,
	OOPS_MOD0,
	OOPS_BADFUNC,
	OOPS_BADCLASS,
	OOPS_BADSYSCALL,
	OOPS_EXIT
};

/* opcodes marked with # in comments take args.
 * args are LONG (4 bytes), unless otherwise staten
 * mov_f has real arg
 * mov_b/w has byte(1)/word(2) args 
 * #(%pc) means realtive to %pc
 */

enum {
	/* op_fault, as well as other undefined instructions cause
	 * OOPS_ILL */
	op_fault,		/* fault */
	/* cause OOPS_HLT */
	op_hlt,			/* hlt */
	op_nop,			/* nop */

	/* calling ... */
	/* run c-code no %d, args in regs/on stack */
	op_syscall,		/* syscall */
	
	/*
	 * alloc space for new instance of class #s, initialize
	 * vptr and class constructor
	 */
	op_new,			/* new #s */

	/* 
	 * allocate bytes for our instance and set up %this
	 * class size should be stored in %d 
	 * first thing to be done in ctor
	 */
	op_alloc,		/* alloc #s */
	
	/*
	 * call %a->%d()
	 * do this: 
	 *	push %this
	 *	mov %a, %this
	 *	push %pc
	 *	mov VIRTUAL_TABLES[id_of(%a)][%d], %pc 
	 */
	op_jsr,			/* jsr */

	/*
	 * same as regular jsr, but class type is forced to arg.
	 * used for prt->Class::func() and ctor/dtor implementation
	 */
	op_jsr_id,		/* jsr_id #s */

	/*
	 * 2b used in dtor, free(%this) 
	 */
	op_die,			/* die */
	
	/*
	 * set up stack frame to # bytes
	 * ->
	 *	push %bp
	 *	mov %sp, %bp
	 *	add %d, %sp
	 */
	op_frame,		/* frame ##? */
	
	/*
	 * return from subroutine
	 * ->
	 *	mov %bp, %sp
	 *	pop %bp
	 *	pop %pc
	 *	pop %this
	 */
	op_rts,			/* rts */
	
	op_clr_d,		/* clr %d */
	op_mov_b,		/* mov.b #d, %d */
	op_mov_w,		/* mov.w #d, %d */
	op_mov_l,		/* mov.l #d, %d */
	op_mov_q,		/* mov.q #lld, %d */
	op_clr_f,		/* clr %f */
	op_mov_f,		/* mov.f #f, %f */
	op_mov_s,		/* mov.s #.25s (%pc), %s */
	op_lda_bp,		/* lda %d(%bp), %a */
	op_lda_this,		/* lda %d(%this), %a */
	op_lda_sys,		/* lda %d(%sys) */
	op_add_a,		/* add %d, %a */
	op_mov_aa,		/* mov (%a), %a */
	op_mov_ad,		/* mov (%a), %d */
	op_mov_af,		/* mov (%a), %f */
	op_mov_as,		/* mov (%a), %s */
	op_mov_da,		/* mov %d, (%a) */
	op_mov_fa,		/* mov %f, (%a) */
	op_mov_sa,		/* mov %s, (%a) */
	op_push_d,		/* push %d (sp+) */
	op_push_a,		/* push %a (sp+) */
	op_push_f,		/* push %f (sp+) */
	op_push_s,		/* push %s (sp+) */
	op_pop_d,		/* pop %d (-sp) */
	op_pop_a,		/* pop %a (-sp) */
	op_pop_f,		/* pop %f (-sp) */
	op_pop_s,		/* pop %s (-sp) */
	op_sub_sp,		/* sub %d, %sp */
	op_neg_d,		/* neg -%d */
	op_neg_f,		/* neg -%f */
	op_inc_d,		/* inc %d++ */
	op_inc_f,		/* inc %f++ */
	op_dec_d,		/* dec %d-- */
	op_dec_f,		/* dec %f-- */
	
	/* stack arithemtic: %reg = (pop arg) [op] %reg */
	op_add_d,		/* add [popd] + %d */
	op_add_f,		/* add [popf] + %f */
	op_add_s,		/* add [pops] + %s */
	op_sub_d,		/* sub [popd] - %d */
	op_sub_f,		/* sub [popf] - %f */
	op_mul_d,		/* mul [popd] * %d */
	op_mul_f,		/* mul [popf] * %f */
	op_div_d,		/* div [popd] / %d */
	op_div_f,		/* div [popf] / %f */
	op_mod_d,		/* mod [popd] % %d */
	op_and_d,		/* and [popd] & %d */
	op_or_d,		/* or [popd] | %d */
	op_eor_d,		/* eor [popd] ^ %d */
	op_shl_d,		/* shl [popd] << %d */
	op_shr_d,		/* shr [popd] >> %d */
	op_not_d,		/* not ~%d */
	op_not,			/* not !%d */
	op_and,			/* and [popd] && %d */
	op_or,			/* or [popd] || %d */
	op_eor,			/* eor [popd] ^^ %d */
	op_conv_fd,		/* conv %f -> %d */
	op_conv_df,		/* conv %d -> %f */
	
	/*
	 * test.XX -> put result of given relations in %d (1/0) 
	 * btrue jump if %d
	 * bfalse jump if !%d 
	 */
	op_teq_d,		/* test.eq [popd] == %d */
	op_tne_d,		/* test.ne [popd] != %d */
	op_tle_d,		/* test.le [popd] <= %d */
	op_tge_d,		/* test.ge [popd] >= %d */
	op_tlt_d,		/* test.lt [popd] < %d */
	op_tgt_d,		/* test.gt [popd] > %d */

	op_teq_f,		/* test.eq [popf] == %f */
	op_tne_f,		/* test.ne [popf] != %f */
	op_tle_f,		/* test.le [popf] <= %f */
	op_tge_f,		/* test.ge [popf] >= %f */
	op_tlt_f,		/* test.lt [popf] < %f */
	op_tgt_f,		/* test.gt [popf] > %f */
	
	op_teq_s,		/* test.eq [pops] == %s */
	op_tne_s,		/* test.ne [pops] != %s */
	op_tle_s,		/* test.le [pops] <= %s */
	op_tge_s,		/* test.ge [pops] >= %s */
	op_tlt_s,		/* test.lt [pops] < %s */
	op_tgt_s,		/* test.gt [pops] > %s */

	op_bra,			/* bra #d (%pc) */
	op_btrue,		/* btrue #d (%pc) */
	op_bfalse,		/* bfalse #d (%pc) */
	
	op_swap_d,		/* swap %d <=> (%sp) */
	op_swap_f,		/* swap %f <=> (%sp) */
	op_swap_s,		/* swap %s <=> (%sp) */
	op_test_f,		/* test %f != 0 -> %d */
	op_mov_dA,		/* mov %d, %a */
	op_mov_Ad,		/* mov %a, %d */
	op_mov_this_a,		/* mov %this, %a */
	op_mov_sys_a,		/* mov %sys, %a */
	op_save_ds,		/* save [%d, %s] */
	op_load_ds,		/* load [%d, %s] */
	op_last
};

